@@ -1,6 +1,6 @@
 /*  GNU ddrescue - Data recovery tool
-    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012
-    Antonio Diaz Diaz.
+    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012,
+    2013 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -38,8 +38,8 @@ int calculate_max_skip_size( const long long isize,
   {
   int skip;
 
-  if( isize > 0 && isize / 100 < Rescuebook::max_skipbs ) skip = isize / 100;
-  else skip = Rescuebook::max_skipbs;
+  if( isize > 0 && isize / 100 < Rb_options::max_skipbs ) skip = isize / 100;
+  else skip = Rb_options::max_skipbs;
   if( skip < hardbs || skip < skipbs ) skip = skipbs;
   return round_up( skip, hardbs );
   }
@@ -84,8 +84,15 @@ int Rescuebook::copy_and_update( const Block & b, const Sblock::Status st,
   int retval = copy_block( b, copied_size, error_size );
   if( !retval )
     {
-    if( copied_size + error_size < b.size() )		// EOF
-      truncate_vector( b.pos() + copied_size + error_size );
+    if( copied_size + error_size < b.size() )			// EOF
+      {
+      if( complete_only ) truncate_domain( b.pos() + copied_size + error_size );
+      else if( !truncate_vector( b.pos() + copied_size + error_size ) )
+        {
+        final_msg( "EOF found before end of logfile" );
+        retval = 1;
+        }
+      }
     if( copied_size > 0 )
       {
       errors += change_chunk_status( Block( b.pos(), copied_size ), Sblock::finished );
@@ -93,7 +100,8 @@ int Rescuebook::copy_and_update( const Block & b, const Sblock::Status st,
       }
     if( error_size > 0 )
       {
-      errors += change_chunk_status( Block( b.pos() + copied_size, error_size ), st );
+      errors += change_chunk_status( Block( b.pos() + copied_size, error_size ),
+                ( error_size > hardbs() ) ? st : Sblock::bad_sector );
       if( iname_ && access( iname_, F_OK ) != 0 )
         {
         final_msg( "input file disappeared" ); final_errno( errno );
@@ -136,8 +144,8 @@ int Rescuebook::copy_non_tried()
       pos = b.end();
       current_status( copying );
       block_found = true;
-      const Sblock::Status st = ( skip_size ? Sblock::bad_sector :
-                                              Sblock::non_trimmed );
+      const Sblock::Status st =
+        ( ( b.size() > hardbs() ) ? Sblock::non_trimmed : Sblock::bad_sector );
       int copied_size = 0, error_size = 0;
       const int retval = copy_and_update( b, st, copied_size, error_size,
                                           "Copying non-tried blocks...",
@@ -152,7 +160,7 @@ int Rescuebook::copy_non_tried()
         {
         if( pos >= 0 && skip_size > 0 )
           {
-          b.pos( pos ); b.size( skip_size ); b.fix_size();
+          b.assign( pos, skip_size ); b.fix_size();
           find_chunk( b, Sblock::non_tried );
           if( pos == b.pos() && b.size() > 0 )
             {
@@ -162,7 +170,7 @@ int Rescuebook::copy_non_tried()
             pos = b.end();
             }
           }
-        if( skip_size < skipbs_ ) skip_size = skipbs_;
+        if( skip_size < skipbs ) skip_size = skipbs;
         else if( skip_size <= max_skip_size / 2 ) skip_size *= 2;
         else skip_size = max_skip_size;
         }
@@ -209,8 +217,8 @@ int Rescuebook::rcopy_non_tried()
       end = b.pos();
       current_status( copying );
       block_found = true;
-      const Sblock::Status st = ( skip_size ? Sblock::bad_sector :
-                                              Sblock::non_trimmed );
+      const Sblock::Status st =
+        ( ( b.size() > hardbs() ) ? Sblock::non_trimmed : Sblock::bad_sector );
       int copied_size = 0, error_size = 0;
       const int retval = copy_and_update( b, st, copied_size, error_size,
                                           "Copying non-tried blocks...",
@@ -235,7 +243,7 @@ int Rescuebook::rcopy_non_tried()
             end = b.pos();
             }
           }
-        if( skip_size < skipbs_ ) skip_size = skipbs_;
+        if( skip_size < skipbs ) skip_size = skipbs;
         else if( skip_size <= max_skip_size / 2 ) skip_size *= 2;
         else skip_size = max_skip_size;
         }
@@ -249,209 +257,188 @@ int Rescuebook::rcopy_non_tried()
 
 
 // Return values: 1 I/O error, 0 OK, -1 interrupted, -2 logfile error.
-// Trim the damaged areas backwards.
+// Trim the damaged areas (largest first) from both edges.
 //
 int Rescuebook::trim_errors()
   {
-  long long end = LLONG_MAX;
+  const char * const msg = "Trimming failed blocks...";
   bool first_post = true;
 
-  while( end > 0 )
+  while( true )
     {
-    long long pos = end - hardbs();
-    if( pos < 0 ) pos = 0;
-    Block b( pos, end - pos );
-    rfind_chunk( b, Sblock::non_trimmed );
-    if( b.size() <= 0 ) break;
-    end = b.pos();
-    current_status( trimming );
-    int copied_size = 0, error_size = 0;
-    const int retval = copy_and_update( b, Sblock::bad_sector, copied_size,
-                                        error_size, "Trimming failed blocks...",
-                                        first_post, false );
-    if( copied_size > 0 ) errsize -= copied_size;
-    if( retval ) return retval;
-    if( error_size > 0 ) error_rate += error_size;
-    if( error_size > 0 && end > 0 )
-      {
-      const int index = find_index( end - 1 );
-      if( index >= 0 && domain().includes( sblock( index ) ) &&
-          sblock( index ).status() == Sblock::non_trimmed )
-        errors += change_chunk_status( sblock( index ), Sblock::non_split );
-      }
-    update_rates();
-    if( !update_logfile( odes_ ) ) return -2;
-    }
-  return 0;
-  }
-
-
-// Return values: 1 I/O error, 0 OK, -1 interrupted, -2 logfile error.
-// Trim the damaged areas in reverse mode, forwards.
-//
-int Rescuebook::rtrim_errors()
-  {
-  long long pos = 0;
-  bool first_post = true;
-
-  while( pos >= 0 )
-    {
-    Block b( pos, hardbs() );
-    find_chunk( b, Sblock::non_trimmed );
-    if( b.size() <= 0 ) break;
-    pos = b.end();
-    current_status( trimming );
-    int copied_size = 0, error_size = 0;
-    const int retval = copy_and_update( b, Sblock::bad_sector, copied_size,
-                                        error_size, "Trimming failed blocks...",
-                                        first_post, true );
-    if( copied_size > 0 ) errsize -= copied_size;
-    if( retval ) return retval;
-    if( error_size > 0 ) error_rate += error_size;
-    if( error_size > 0 && pos >= 0 )
-      {
-      const int index = find_index( pos );
-      if( index >= 0 && domain().includes( sblock( index ) ) &&
-          sblock( index ).status() == Sblock::non_trimmed )
-        errors += change_chunk_status( sblock( index ), Sblock::non_split );
-      }
-    update_rates();
-    if( !update_logfile( odes_ ) ) return -2;
-    }
-  return 0;
-  }
-
-
-// Return values: 1 I/O error, 0 OK, -1 interrupted, -2 logfile error.
-// Try to read the damaged areas, splitting them into smaller pieces.
-//
-int Rescuebook::split_errors()
-  {
-  bool first_post = true;
-
-  for( bool first_pass = true; ; first_pass = false )
-    {
-    long long pos = 0;
-    int error_counter = 0;
-    bool block_found = false;
-
-    if( first_pass && current_status() == splitting &&
-        domain().includes( current_pos() ) )
-      {
-      Block b( current_pos(), 1 );
-      find_chunk( b, Sblock::non_split );
-      if( b.size() > 0 ) pos = b.pos();		// resume
-      }
-
+    const int index = find_largest_sblock( Sblock::non_trimmed );
+    if( index < 0 ) break;				// no more blocks
+    const Block block = sblock( index );
+    long long pos = block.pos();
     while( pos >= 0 )
       {
       Block b( pos, hardbs() );
-      find_chunk( b, Sblock::non_split );
-      if( b.size() <= 0 ) break;
-      if( pos != b.pos() ) error_counter = 0;	// reset cntr on block change
+      find_chunk( b, Sblock::non_trimmed );
+      if( pos != b.pos() || b.size() <= 0 ) break;	// block change
       pos = b.end();
-      current_status( splitting );
-      block_found = true;
+      current_status( trimming );
       int copied_size = 0, error_size = 0;
       const int retval = copy_and_update( b, Sblock::bad_sector, copied_size,
-                                          error_size, "Splitting failed blocks...",
-                                          first_post, true );
+                                          error_size, msg, first_post, true );
+      if( copied_size > 0 ) errsize -= copied_size;
+      if( retval ) return retval;
+      if( error_size > 0 ) { error_rate += error_size; pos = -1; }
+      update_rates();
+      if( !update_logfile( odes_ ) ) return -2;
+      }
+    long long end = block.end();
+    while( end > 0 )
+      {
+      pos = end - hardbs();
+      if( pos < 0 ) pos = 0;
+      Block b( pos, end - pos );
+      rfind_chunk( b, Sblock::non_trimmed );
+      if( pos != b.pos() || b.size() <= 0 ) break;	// block change
+      end = b.pos();
+      current_status( trimming );
+      int copied_size = 0, error_size = 0;
+      const int retval = copy_and_update( b, Sblock::bad_sector, copied_size,
+                                          error_size, msg, first_post, false );
       if( copied_size > 0 ) errsize -= copied_size;
       if( retval ) return retval;
       if( error_size > 0 ) error_rate += error_size;
-      if( error_size <= 0 ) error_counter = 0;
-      else if( pos >= 0 && ++error_counter >= 8 )
-        {			// skip after enough consecutive errors
-        error_counter = 0;
-        const int index = find_index( pos );
-        if( index >= 0 )
-          {
-          const Sblock & sb = sblock( index );
-          if( sb.status() == Sblock::non_split && sb.size() >= skipbs_ )
-            {
-            if( sb.size() >= 2 * skipbs_ )	// split
-              {
-              if( first_pass || sblocks() < 1000 )
-                pos += ( sb.size() / ( 2 * hardbs() ) ) * hardbs();
-              else pos = sb.end() - skipbs_;
-              }
-            else pos = sb.end();		// skip
-            }
-          }
+      if( error_size > 0 && end > 0 )
+        {
+        const int index = find_index( end - 1 );
+        if( index >= 0 && domain().includes( sblock( index ) ) &&
+            sblock( index ).status() == Sblock::non_trimmed )
+          errors += change_chunk_status( sblock( index ), Sblock::non_split );
+        end = -1;
         }
       update_rates();
       if( !update_logfile( odes_ ) ) return -2;
       }
-    if( !block_found ) break;
     }
   return 0;
   }
 
 
 // Return values: 1 I/O error, 0 OK, -1 interrupted, -2 logfile error.
-// Try to read the damaged areas in reverse mode, splitting them into
-// smaller pieces.
+// Split the damaged areas (largest first).
+// Then read the remaining small areas sequentially.
 //
-int Rescuebook::rsplit_errors()
+int Rescuebook::split_errors( const bool reverse )
   {
+  const char * const msg = "Splitting failed blocks...";
   bool first_post = true;
 
-  for( bool first_pass = true; ; first_pass = false )
+  while( true )
+    {
+    if( sblocks() < max_logfile_size )			// split largest block
+      {
+      const int index = find_largest_sblock( Sblock::non_split );
+      if( index < 0 ) break;				// no more blocks
+      const Block block = sblock( index );
+      if( block.size() / hardbs() < 5 ) break;		// no more large blocks
+      const long long midpos =
+        block.pos() + ( ( block.size() / ( 2 * hardbs() ) ) * hardbs() );
+      long long pos = midpos;
+      while( pos >= 0 )
+        {
+        Block b( pos, hardbs() );
+        find_chunk( b, Sblock::non_split );
+        if( pos != b.pos() || b.size() <= 0 ) break;	// block change
+        pos = b.end();
+        current_status( splitting );
+        int copied_size = 0, error_size = 0;
+        const int retval = copy_and_update( b, Sblock::bad_sector, copied_size,
+                                            error_size, msg, first_post, true );
+        if( copied_size > 0 ) errsize -= copied_size;
+        if( retval ) return retval;
+        if( error_size > 0 ) { error_rate += error_size; pos = -1; }
+        update_rates();
+        if( !update_logfile( odes_ ) ) return -2;
+        }
+      long long end = midpos;
+      while( end > 0 )
+        {
+        pos = end - hardbs();
+        if( pos < 0 ) pos = 0;
+        Block b( pos, end - pos );
+        rfind_chunk( b, Sblock::non_split );
+        if( pos != b.pos() || b.size() <= 0 ) break;	// block change
+        end = b.pos();
+        current_status( splitting );
+        int copied_size = 0, error_size = 0;
+        const int retval = copy_and_update( b, Sblock::bad_sector, copied_size,
+                                            error_size, msg, first_post, true );
+        if( copied_size > 0 ) errsize -= copied_size;
+        if( retval ) return retval;
+        if( error_size > 0 ) { error_rate += error_size; end = -1; }
+        update_rates();
+        if( !update_logfile( odes_ ) ) return -2;
+        }
+      }
+    else			// logfile is full; read smallest block
+      {
+      const int index = find_smallest_sblock( Sblock::non_split );
+      if( index < 0 ) break;				// no more blocks
+      const Block block = sblock( index );
+      long long pos = block.pos();
+      while( pos >= 0 )
+        {
+        Block b( pos, hardbs() );
+        find_chunk( b, Sblock::non_split );
+        if( pos != b.pos() || b.size() <= 0 ) break;	// block change
+        pos = b.end();
+        current_status( splitting );
+        int copied_size = 0, error_size = 0;
+        const int retval = copy_and_update( b, Sblock::bad_sector, copied_size,
+                                            error_size, msg, first_post, true );
+        if( copied_size > 0 ) errsize -= copied_size;
+        if( retval ) return retval;
+        if( error_size > 0 ) error_rate += error_size;
+        update_rates();
+        if( !update_logfile( odes_ ) ) return -2;
+        }
+      }
+    }
+  if( !reverse )		// read the remaining small areas forwards
+    {
+    long long pos = 0;
+    while( pos >= 0 )
+      {
+      Block b( pos, hardbs() );
+      find_chunk( b, Sblock::non_split );
+      if( b.size() <= 0 ) break;			// no more blocks
+      pos = b.end();
+      current_status( splitting );
+      int copied_size = 0, error_size = 0;
+      const int retval = copy_and_update( b, Sblock::bad_sector, copied_size,
+                                          error_size, msg, first_post, true );
+      if( copied_size > 0 ) errsize -= copied_size;
+      if( retval ) return retval;
+      if( error_size > 0 ) error_rate += error_size;
+      update_rates();
+      if( !update_logfile( odes_ ) ) return -2;
+      }
+    }
+  else				// read the remaining small areas backwards
     {
     long long end = LLONG_MAX;
-    int error_counter = 0;
-    bool block_found = false;
-
-    if( first_pass && current_status() == splitting &&
-        domain().includes( current_pos() - 1 ) )
-      {
-      Block b( current_pos() - 1, 1 );
-      rfind_chunk( b, Sblock::non_split );
-      if( b.size() > 0 ) end = b.end();		// resume
-      }
-
     while( end > 0 )
       {
       long long pos = end - hardbs();
       if( pos < 0 ) pos = 0;
       Block b( pos, end - pos );
       rfind_chunk( b, Sblock::non_split );
-      if( b.size() <= 0 ) break;
-      if( pos != b.pos() ) error_counter = 0;	// reset cntr on block change
+      if( b.size() <= 0 ) break;			// no more blocks
       end = b.pos();
       current_status( splitting );
-      block_found = true;
       int copied_size = 0, error_size = 0;
       const int retval = copy_and_update( b, Sblock::bad_sector, copied_size,
-                                          error_size, "Splitting failed blocks...",
-                                          first_post, false );
+                                          error_size, msg, first_post, true );
       if( copied_size > 0 ) errsize -= copied_size;
       if( retval ) return retval;
       if( error_size > 0 ) error_rate += error_size;
-      if( error_size <= 0 ) error_counter = 0;
-      else if( end > 0 && ++error_counter >= 8 )
-        {			// skip after enough consecutive errors
-        error_counter = 0;
-        const int index = find_index( end - 1 );
-        if( index >= 0 )
-          {
-          const Sblock & sb = sblock( index );
-          if( sb.status() == Sblock::non_split && sb.size() >= skipbs_ )
-            {
-            if( sb.size() >= 2 * skipbs_ )	// split
-              {
-              if( first_pass || sblocks() < 1000 )
-                end -= ( sb.size() / ( 2 * hardbs() ) ) * hardbs();
-              else end = sb.pos() + skipbs_;
-              }
-            else end = sb.pos();		// skip
-            }
-          }
-        }
       update_rates();
       if( !update_logfile( odes_ ) ) return -2;
       }
-    if( !block_found ) break;
     }
   return 0;
   }
@@ -465,7 +452,7 @@ int Rescuebook::copy_errors()
   char msgbuf[80] = "Retrying bad sectors... Retry ";
   const int msglen = std::strlen( msgbuf );
 
-  for( int retry = 1; max_retries_ < 0 || retry <= max_retries_; ++retry )
+  for( int retry = 1; max_retries < 0 || retry <= max_retries; ++retry )
     {
     long long pos = 0;
     bool first_post = true, block_found = false;
@@ -510,7 +497,7 @@ int Rescuebook::rcopy_errors()
   char msgbuf[80] = "Retrying bad sectors... Retry ";
   const int msglen = std::strlen( msgbuf );
 
-  for( int retry = 1; max_retries_ < 0 || retry <= max_retries_; ++retry )
+  for( int retry = 1; max_retries < 0 || retry <= max_retries; ++retry )
     {
     long long end = LLONG_MAX;
     bool first_post = true, block_found = false;
@@ -550,32 +537,20 @@ int Rescuebook::rcopy_errors()
 
 
 Rescuebook::Rescuebook( const long long offset, const long long isize,
-                        const long long max_error_rate,
-                        const long long min_outfile_size,
-                        const long long min_read_rate, Domain & dom,
+                        Domain & dom, const Rb_options & rb_opts,
                         const char * const iname, const char * const logname,
-                        const long timeout, const int cluster, const int hardbs,
-                        const int skipbs, const int max_errors,
-                        const int max_retries, const bool complete_only,
-                        const bool new_errors_only, const bool nosplit,
-                        const bool retrim, const bool sparse,
-                        const bool synchronous, const bool try_again )
-  : Logbook( offset, isize, dom, logname, cluster, hardbs, complete_only ),
-    max_error_rate_( max_error_rate ),
-    min_outfile_size_( min_outfile_size ),
+                        const int cluster, const int hardbs,
+                        const bool synchronous )
+  : Logbook( offset, isize, dom, logname, cluster, hardbs, rb_opts.complete_only ),
+    Rb_options( rb_opts ),
     error_rate( 0 ),
-    min_read_rate_( min_read_rate ),
     sparse_size( sparse ? 0 : -1 ),
     recsize( 0 ),
     errsize( 0 ),
-    timeout_( timeout ),
     iname_( ( access( iname, F_OK ) == 0 ) ? iname : 0 ),
-    max_retries_( max_retries ),
-    skipbs_( skipbs ),
     max_skip_size( calculate_max_skip_size( isize, hardbs, skipbs ) ),
-    max_errors_( max_errors ),
     e_code( 0 ),
-    nosplit_( nosplit ), synchronous_( synchronous ),
+    synchronous_( synchronous ),
     a_rate( 0 ), c_rate( 0 ), first_size( 0 ), last_size( 0 ),
     last_ipos( 0 ), t0( 0 ), t1( 0 ), ts( 0 ), oldlen( 0 ),
     rates_updated( false )
@@ -601,7 +576,7 @@ Rescuebook::Rescuebook( const long long offset, const long long isize,
         change_sblock_status( index, Sblock::non_tried );
       }
   count_errors();
-  if( new_errors_only ) max_errors_ += errors;
+  if( new_errors_only ) max_errors += errors;
   }
 
 
@@ -653,12 +628,12 @@ int Rescuebook::do_rescue( const int ides, const int odes, const bool reverse )
   if( copy_pending && !errors_or_timeout() )
     retval = ( reverse ? rcopy_non_tried() : copy_non_tried() );
   if( !retval && trim_pending && !errors_or_timeout() )
-    retval = ( reverse ? rtrim_errors() : trim_errors() );
-  if( !retval && split_pending && !nosplit_ && !errors_or_timeout() )
-    retval = ( reverse ? rsplit_errors() : split_errors() );
-  if( !retval && max_retries_ != 0 && !errors_or_timeout() )
+    retval = trim_errors();
+  if( !retval && split_pending && !nosplit && !errors_or_timeout() )
+    retval = split_errors( reverse );
+  if( !retval && max_retries != 0 && !errors_or_timeout() )
     retval = ( reverse ? rcopy_errors() : copy_errors() );
-  if( !rates_updated ) update_rates( true );	// force update of rates
+  if( !rates_updated ) update_rates( true );	// force update of e_code
   show_status( -1, (retval ? 0 : "Finished"), true );
   if( !retval && errors_or_timeout() ) retval = 1;
   if( verbosity >= 0 )
